/**
 * Hunter Adams (vha3@cornell.edu)
 * 
 * HARDWARE CONNECTIONS
 *  - GPIO 16 ---> VGA Hsync
 *  - GPIO 17 ---> VGA Vsync
 *  - GPIO 18 ---> 470 ohm resistor ---> VGA Green 
 *  - GPIO 19 ---> 330 ohm resistor ---> VGA Green
 *  - GPIO 20 ---> 330 ohm resistor ---> VGA Blue
 *  - GPIO 21 ---> 330 ohm resistor ---> VGA Red
 *  - RP2040 GND ---> VGA GND
 * 
 * RESOURCES USED
 *  - PIO state machines 0, 1, and 2 on PIO instance 0
 *  - DMA channels 0, 1, 2, and 3
 *  - 153.6 kBytes of RAM (for pixel color data)
 *
 * Protothreads v1.3
 * Threads:
 * core 0:
 * Graphics demo
 * blink LED25 
 * core 1:
 * Toggle gpio 4 
 * Serial i/o 
 */
// ==========================================
// === VGA graphics library
// ==========================================
#include "vga16_graphics.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "pico/stdlib.h"
#include "hardware/pio.h"
#include "hardware/dma.h"

// ==========================================
// === protothreads globals
// ==========================================
#include "hardware/sync.h"
#include "hardware/timer.h"
#include "pico/multicore.h"
#include "string.h"
// protothreads header
#include "pt_cornell_rp2040_v1_3.h"

char user_string[40] = "Type up to 40 characters" ;
int new_str = 1;

// ==========================================
// === fixed point s15x16
// ==========================================
// s15x16 fixed point macros ==
// == resolution 2^-16 = 1.5e-5
// == dynamic range is 32767/-32768
typedef signed int s15x16;
#define muls15x16(a,b) ((s15x16)(((( signed long long )(a))*(( signed long long )(b)))>>16)) //multiply two fixed 16:16
#define float_to_s15x16(a) ((s15x16)((a)*65536.0)) // 2^16
#define s15x16_to_float(a) ((float)(a)/65536.0)
#define s15x16_to_int(a)    ((int)((a)>>16))
#define int_to_s15x16(a)    ((s15x16)((a)<<16))
#define divs15x16(a,b) ((s15x16)((((signed long long)(a)<<16)/(b)))) 
#define abss15x16(a) abs(a)

// ==========================================
// === fixed point s4x27
// ==========================================
// s15x16 fixed point macros ==
// == resolution 2^-16 = 1.5e-5
// == dynamic range is 32767/-32768
typedef signed int s4x27;
#define two27 134217728
#define muls4x27(a,b) ((s4x27)(((( signed long long )(a))*(( signed long long )(b)))>>27)) //multiply two fixed 16:16
#define float_to_s4x27(a) ((s4x27)((a)*two27)) // 2^16
#define s4x27_to_float(a) ((float)(a)/two27)
#define s4x27_to_int(a)    ((int)((a)>>27))
#define int_to_s4x27(a)    ((s4x27)((a)<<27))
#define divs4x27(a,b) ((s4x27)((((signed long long)(a)<<27)/(b)))) 
#define abss4x27(a) abs(a)

// =============================================
// fp16 format
// =============================================
// bit 15 is sign
// 5 bits 14-10 exponent
// 10 bits 9-0 mantissa (with one implied)
// if exp_offset is set to 15, then
// represent one as 0|01111|0000000000
// a zero exponent implies number=0
typedef unsigned short fp16V ;
// Max exp is 31, min is 0
// setting 15 gives a range of +/-1e5 and smallest value ~ 1e-5
#define exp_offset 28
// derive the max/min number bits
#define maxpos 0x7fff
#define maxneg 0xffff
// get the actual max/min for a given exp_offset
float maxnum ;
float minnum ;

//
// =============================================
// type conversions
// =============================================
fp16V IEEE_to_fp16V(float a){
  fp16V b ;
  if(a==0) return 0 ;
  if(fabs(a)>maxnum) a = (a>0)? maxnum : -maxnum ;
  if(fabs(a)<minnum) a = (a>0)? minnum : -minnum ;
  int ia = (*(int*)&a); 
  int sgn = (ia>>31) ;
  int exponent = ((ia>>23) & 0xff) - 127 + exp_offset ; // -127 + 15
  int mantissa = ia & 0x7fffff ;
  // debugging
 // printf("%f %d %x %x\n\r", a, sgn, exponent, mantissa>>13);
  b = (sgn<<15) | ((exponent)<<10)  | (mantissa>>13)  ; // >>23
  return b ;
}
// 
float fp16V_to_IEEE(fp16V a){
  int ia = (int) a ;
  int sgn = (ia>>15) & 0x01 ;
  int exponent = (ia>>10) & 0x1f ;
  if (exponent==0) return 0.0 ;
  int mantissa = ia & 0x3ff ;
  float result = pow(2.0, (float) (exponent - exp_offset));
  result = (1.0+(((float)mantissa) / 1024.0)) * result;
  if(sgn) result = result * (-1);
  return result;
}

// integer conversions
#define int_to_fp16V(a) (IEEE_to_fp16V((float)(a)))
#define fp16V_to_int(a) ((int)fp16V_to_IEEE(a))

// =============================================
// negate and absolute
// =============================================
#define fp16V_neg(a) (((a) & 0x8000)? (a) & 0x7fff : (a) | 0x8000)
#define fp16V_abs(a) ((a) & 0x7fff)

// =============================================
// add/subtract
// =============================================
fp16V fp16V_add(fp16V a, fp16V b){
  // isolate mantissas and add one
  uint32_t aFrac = (a & 0x3ff) | 0x400 ;
  uint32_t bFrac = (b & 0x3ff) | 0x400 ;
  // isolate the exponents
  int aExp = (a >> 10) & 0x1f ;
  int bExp = (b >> 10) & 0x1f ;
  int expDiff = aExp - bExp ;
  // isolate the signs
  uint32_t aSign = (a >> 15) & 0x01 ;
  uint32_t bSign = (b >> 15) & 0x01 ;
  // debugging
  //printf("%x %x %x %x %x %x \n\r", aFrac, bFrac, aExp, bExp, aSign, bSign);

  // add_sub flag is 1 for subtract
  uint32_t add_sub = aSign ^ bSign ;
  //
  if (aExp==0 | expDiff<-10) return b ;  //10
  if (bExp==0 | expDiff>10) return a ;// 10
  //
  if (expDiff>0){ // a bigger
    // add_sub is 1 for subtract
    if(add_sub){ // subtract
      // result is between 1 and 0.0
      uint32_t sum = aFrac - (bFrac>>expDiff) ;
      while (sum < 0x400) {
        aExp--;
        sum <<= 1;
      }
      return (aSign<<15) | (aExp<<10) | (sum & 0x3ff)  ;
    }
    else { //add
      // result is between 1 and 2
      uint32_t sum = aFrac + (bFrac>>expDiff) ;
      if (sum >= 0x800) {
        aExp++;
        sum >>= 1;
      }
      return (aSign<<15) | (aExp<<10) | (sum & 0x3ff)  ;
    }
  }
  //
  else if (expDiff<0){ // b bigger
    // add_sub is 1 for subtract
    if(add_sub){ // subtract
      uint32_t sum = bFrac - (aFrac>>-expDiff) ;
      while (sum < 0x400) {
        bExp--;
        sum <<= 1;
      }
      return (bSign<<15) | (bExp<<10) | (sum & 0x3ff)  ;
    }
    else { //add
      uint32_t sum = bFrac + (aFrac>>-expDiff) ;
      if (sum >= 0x800) {
        bExp++;
        sum >>= 1;
      }
      return (bSign<<15) | (bExp<<10) | (sum & 0x3ff)  ;
    }
  }
  else if (expDiff==0){ // 
    // add_sub is 1 for subtract
    if(add_sub){ // subtract
      // figure out which mantissa is bigger
      // compute bigger - smaller : result<1
      // normalize
      if (bFrac == aFrac) return 0 ;
      uint32_t sum = (bFrac > aFrac)? bFrac - aFrac : aFrac - bFrac ;
      while (sum < 0x400) {
        bExp--;
        sum <<= 1;
        if (bExp == 0) return 0 ;
      }
      int Sign = (bFrac > aFrac)? bSign : aSign ;
      return (Sign<<15) | (bExp<<10) | (sum & 0x3ff)  ;
    }
    else { //add
      uint32_t sum = bFrac + aFrac ;
      // sum is between 2 and 4
      bExp++;
      sum >>= 1;
      return (bSign<<15) | (bExp<<10) | (sum & 0x3ff)  ;
    }
  }
}

// ==================
fp16V fp16V_sub(fp16V a, fp16V b){
  fp16V diff = fp16V_add(a, fp16V_neg(b)) ;
  return diff ;
}


// =============================================
// multiply
// =============================================
fp16V fp16V_mul(fp16V a, fp16V b){
  // isolate mantissas and add one
  uint32_t aFrac = (a & 0x3ff) | 0x400 ;
  uint32_t bFrac = (b & 0x3ff) | 0x400 ;
  // isolate the exponents
  int aExp = (a >> 10) & 0x1f ;
  int bExp = (b >> 10) & 0x1f ;
  // isolate the signs
  uint32_t aSign = (a >> 15) & 0x01 ;
  uint32_t bSign = (b >> 15) & 0x01 ;
  // add exponents an subtract offset
  int expSum = aExp + bExp - exp_offset ;
  // get output sign
  uint32_t Sign = aSign ^ bSign ;
  //
  if (aExp==0 | bExp==0) return 0 ;
  if(expSum > 31) return maxnum ;
  if(aExp + bExp < exp_offset) return 0 ;

  // range is 1 to 4
  uint32_t product = (aFrac * bFrac) >> 10 ;
  if (product >= 0x800) {
    expSum++;
    product >>= 1;
    if(expSum > 31) return maxnum ;
  }
  return (Sign<<15) | (expSum<<10) | (product & 0x3ff) ;
}


// =============================================
// divide
// =============================================
fp16V fp16V_div(fp16V a, fp16V b){
 // isolate mantissas and add one
  uint32_t aFrac = (a & 0x3ff) | 0x400 ;
  uint32_t bFrac = (b & 0x3ff) | 0x400 ;
  // isolate the exponents
  int aExp = (a >> 10) & 0x1f ;
  int bExp = (b >> 10) & 0x1f ;
  // isolate the signs
  uint32_t aSign = (a >> 15) & 0x01 ;
  uint32_t bSign = (b >> 15) & 0x01 ;
  // sub exponents an add offset
  int expDiff = aExp - bExp + exp_offset ;
  // get output sign
  uint32_t Sign = aSign ^ bSign ;
  //
  if (aExp==0 | bExp==0) return 0 ;
  if(expDiff< 1) return 0 ;
  if(aExp - bExp > exp_offset) return 0 ;

  // range is 2 to .5
  uint32_t ratio = (aFrac<<10) / bFrac  ;
  if (ratio < 0x400) {
    expDiff--;
    ratio <<= 1;
    if(expDiff < 1) return 0 ;
  }
  return (Sign<<15) | (expDiff<<10) | (ratio & 0x3ff) ;
}


// =============================================
// floating compare
// =============================================
// returns 1 if a >= b otherwise zero
char fp16V_cmp(fp16V a, fp16V b){
 // isolate mantissas and add one
  uint32_t aFrac = (a & 0x3ff) | 0x400 ;
  uint32_t bFrac = (b & 0x3ff) | 0x400 ;
  // isolate the exponents
  int aExp = (a >> 10) & 0x1f ;
  int bExp = (b >> 10) & 0x1f ;
  // isolate the signs
  uint32_t aSign = (a >> 15) & 0x01 ;
  uint32_t bSign = (b >> 15) & 0x01 ;
   // a neg b pos
  if (aSign > bSign) return 0 ; 
  // b neg a pos
  else if (aSign < bSign) return 1 ;
  // both pos
  else if (!aSign) {
    if ((a >= b)) return 1;
    else return 0 ;
  }
  // both neg
  else if (aSign) {
    if ((a <= b)) return 1;
    else return 0 ;
  }   
}


// =============================================
// inverse sqrt
// =============================================
// see https://en.wikipedia.org/wiki/Fast_inverse_square_root
// particularly the section "First approximation of the result"
// The 'magic number' is  1.5 * 2^(num_mantissa_bits) * ((exp_offset)-0.045)
// For fp16; exp offset = 15 ;  num mantissa bits = 10
fp16V fp16V_invsqrt( fp16V number ){
	int i;
	fp16V x2, y;
	const fp16V threehalfs = IEEE_to_fp16V(1.5);
  const fp16V onehalfs = IEEE_to_fp16V(0.5);
  const fp16V small = IEEE_to_fp16V(2*minnum);

  if(!fp16V_cmp(number, small))  return 0 ;
  
	x2 = fp16V_mul(number, onehalfs);
	y  = number;
	i  = * ( int * ) &y;       // evil floating point bit level hacking
                             // exp offset = 15 ;  num mantissa bits = 10
	i  = 22971 - ( i >> 1 );    // 1.5 * 2^(num mantissa bits) * ((exp offset)-0.045)  // what the f?
	y  = (short) i ;           //* ( fp16 * ) &i;
  // now one Newton iteration
	y  = fp16V_mul(y, fp16V_add(threehalfs, fp16V_neg((fp16V_mul(x2,fp16V_mul(y,y))))));   // 1st iteration
//	y  = y * ( threehalfs - ( x2 * y * y ) );   // 2nd iteration, this can be removed
	return y;
}

// =============================================
// floating shift
// =============================================
// positive b means shift left, negaive is shift right
fp16V fp16V_shift( fp16V a, signed char b ){
  // isolate the exponent
  int aExp = (a >> 10) & 0x1f ;
   aExp = (aExp + b) & 0x1f ;
   // mask out exponent and OR  in new one
   a = (a & 0x83ff ) | (aExp << 10) ;
   return a ;
}

// =============================================
// definitions to make FP arithmetic more compact
// =============================================
#define add fp16V_add
#define sub fp16V_sub
#define cmp fp16V_cmp
#define sft fp16V_shift
#define isqr fp16V_invsqrt
#define mul fp16V_mul
#define div fp16V_div
#define neg fp16V_neg
#define abs fp16V_abs

static float t, d_m=0.2, k_m=2;
static float kv1_fp32, kv2_fp32, mv1_fp32, mv2_fp32, v1t_fp32, v2t_fp32, v1_fp32, v2_fp32 ;
static float v1_exact = 1.0, v2_exact = 0;
static float w; 


static s15x16  d_m_s15x16, k_m_s15x16 ;
static s15x16 kv1_s15x16, kv2_s15x16, mv1_s15x16, mv2_s15x16, v1t_s15x16, v2t_s15x16, v1_s15x16, v2_s15x16 ;

static s4x27  d_m_s4x27, k_m_s4x27 ;
static s4x27 kv1_s4x27, kv2_s4x27, mv1_s4x27, mv2_s4x27, v1t_s4x27, v2t_s4x27, v1_s4x27, v2_s4x27 ;
    //
static fp16V d_m_fp16V, k_m_fp16V ;
static fp16V kv1_fp16V, kv2_fp16V, mv1_fp16V, mv2_fp16V, v1t_fp16V, v2t_fp16V, v1_fp16V, v2_fp16V ;

int log_dt = 5 ;
float dt ;
s15x16 dt_s15x16 ;
s4x27 dt_s4x27 ;

bool show_exact, show_fp32, show_fp16V, show_s15x16, show_s4x27 ;
float error_gain = 10 ;
int h_scale = 1, x_pos = 0;


// ==================================================
// === graphics demo -- RUNNING on core 0
// ==================================================
static PT_THREAD (protothread_graphics(struct pt *pt)) {
    PT_BEGIN(pt);
    // the protothreads interval timer
    PT_INTERVAL_INIT() ;

    // init a couple of cnstants for fp16V
    maxnum = pow(2, 31-exp_offset+1) ;
    minnum = pow(2, -exp_offset+1) ;

    // Draw some filled rectangles
    fillRect(64, 0, 176, 50, BLUE); // blue box
    //fillRect(250, 0, 176, 50, DARK_ORANGE); // red box
    fillRect(435, 0, 176, 50, CYAN); // green box

    // Write some text
    setTextColor(WHITE) ;
    setCursor(65, 0) ;
    setTextSize(1) ;
    writeString("Raspberry Pi Pico") ;
    setCursor(65, 10) ;
    writeString("FP demo") ;
    setCursor(65, 20) ;
    writeString("Hunter Adams") ;
    setCursor(65, 30) ;
    writeString("Bruce Land") ;
    setCursor(65, 40) ;
    writeString("ece4760 Cornell") ;
    //
    setCursor(445, 10) ;
    setTextColor(BLACK) ;
    setTextSize(1) ;
    writeString("Protothreads rp2040 v1.3") ;
    setCursor(445, 20) ;
    writeString("VGA 4-bit color 640x480") ;
    setCursor(445, 30) ;
    writeString("Tested on pico and picoW") ;
    //
    setCursor(250, 20) ;
    setTextColor2(WHITE, BLACK) ;
    writeStringBig("Floating Point Test") ;
    setCursor(250, 37) ;
    setTextColor2(WHITE, BLACK) ;
    writeStringBold("Compare fp32, fp16, s15x16") ;
    

    // uaw Heun integrator to simulate 2nd order linear
    /*
    Form the slope for each variable at time=n
    kv1 = f1(v1(n), v2(n))
    kv2 = f2(v1(n), v2(n))
    Estimate one step forward for each variable
    v1t = v1 + dt*kv1
    v2t = v2 + dt*kv2
    Form the slope for each variable at time=n+1
    mv1 = f1(v1t, v2t)
    mv2 = f2(v1t, v2t)
    Average the two estimates to get the new values of V.
    v1(n+1) = v1(n) + dt*(kv1 + mv1)/2
    v2(n+1) = v2(n) + dt*(kv2 + mv2)/2
	*/
    
    
    // timing operations
    int64_t start_time ;
    static char video_string[64] ;
    
    while(true) {

        PT_YIELD_UNTIL(pt, new_str);
        new_str = 0;
        
        // erase screen
        fillRect(0, 70, 640, 410, BLACK) ;
        // print parameters
        sprintf(video_string, "k/m=%5.2f", k_m);
        setCursor(70, 70) ;
        setTextColor2(WHITE, BLACK) ;
        writeStringBig(video_string) ;

        sprintf(video_string, "d/m=%5.2f", d_m);
        setCursor(170, 70) ;
        setTextColor2(WHITE, BLACK) ;
        writeStringBig(video_string) ;

        sprintf(video_string, "log_dt=%2d", log_dt);
        setCursor(270, 70) ;
        setTextColor2(WHITE, BLACK) ;
        writeStringBig(video_string) ;

        sprintf(video_string, "h_scale=%2d", h_scale);
        setCursor(370, 70) ;
        setTextColor2(WHITE, BLACK) ;
        writeStringBig(video_string) ;

        sprintf(video_string, "error_gain=%3.0f", error_gain);
        setCursor(470, 70) ;
        setTextColor2(WHITE, BLACK) ;
        writeStringBig(video_string) ;

        //
         w = sqrt(k_m) ;
        d_m_s15x16=float_to_s15x16(d_m);
        k_m_s15x16=float_to_s15x16(k_m);
      
        d_m_s4x27 =float_to_s4x27(d_m);
        k_m_s4x27=float_to_s4x27(k_m);

        d_m_fp16V=IEEE_to_fp16V(-d_m) ;
        k_m_fp16V=IEEE_to_fp16V(-k_m) ;
      
        dt=pow(2, -log_dt) ;
        dt_s15x16 = float_to_s15x16(dt) ;
        dt_s4x27 = float_to_s4x27(dt) ;
        //fp16V dt_fp16V = IEEE_to_fp16V(dt) ;

        v1_exact = 1.0; v2_exact = 0;
        v1_fp32 = 1.0; v2_fp32 = 0;
        v1_s15x16 = float_to_s15x16(1.0); v2_s15x16 = float_to_s15x16(0);
        v1_s4x27 = float_to_s4x27(1.0); v2_s4x27 = float_to_s4x27(0);
        v1_fp16V = IEEE_to_fp16V(1.0) ;  v2_fp16V = IEEE_to_fp16V(0.0) ;
        t = 0. ;
        x_pos = 0 ;

        // compute  exact solution
        for(int i=0; i<(639 * h_scale); i++){

          //  draw full curves
          #define gain (100)
          if((i % h_scale) == 0){
            drawPixel(x_pos, (int)(-v1_exact*gain + 200), WHITE);
            if(show_fp32) drawPixel(x_pos, (int)(-v1_fp32*gain + 200), RED); 
            if(show_s15x16) drawPixel(x_pos, s15x16_to_int((muls15x16(-v1_s15x16, int_to_s15x16(gain))+int_to_s15x16(200))), LIGHT_BLUE);
            if(show_fp16V) drawPixel(x_pos, fp16V_to_IEEE(neg(v1_fp16V)) * gain + 200, GREEN);
            if(show_s4x27) drawPixel(x_pos, (int)((s4x27_to_float(-v1_s4x27) * gain + 200)), BLUE);
            // draw errors
            float e_gain =  (gain*error_gain) ;
            float error_fp32 = v1_fp32 - v1_exact ;
            if(show_fp32) drawPixel(x_pos, (int)(error_fp32*e_gain + 350), RED);

            float error_s15x16 = s15x16_to_float(v1_s15x16) - v1_exact ;
            if(show_s15x16) drawPixel(x_pos, (int)(error_s15x16*e_gain+350), LIGHT_BLUE);

            float error_s4x27 = s4x27_to_float(v1_s4x27) - v1_exact ;
            if(show_s4x27) drawPixel(x_pos, (int)(error_s4x27*e_gain+350), BLUE);

            float error_fp16V = fp16V_to_IEEE(v1_fp16V) - v1_exact ;
            if(show_fp16V) drawPixel(x_pos, (int)(error_fp16V*e_gain+350), GREEN);

            x_pos++ ;
          }

          // exact solution
          t = t + dt ;
          //v1_exact = exp(-d_m/2 * (t)) * cos(sqrt(k_m - (d_m*d_m)) * (t)) ; // - (d_m*d_m)/4
          v1_exact = exp(-d_m/2 * t) * (cos(w*t) + d_m/(2*w) * sin(w*t)); // - (d_m*d_m)/4
    
          //
          // compute DDA in fp32
          //start_time = PT_GET_TIME_usec();

          kv1_fp32 = v2_fp32;
          kv2_fp32 = -k_m*v1_fp32 - d_m*v2_fp32 ;

          v1t_fp32 = v1_fp32 + dt * kv1_fp32;
          v2t_fp32 = v2_fp32 + dt * kv2_fp32;

          mv1_fp32 = v2t_fp32;
          mv2_fp32 = -k_m*v1t_fp32 - d_m*v2t_fp32 ;

          v1_fp32 = v1_fp32 + dt * (kv1_fp32 + mv1_fp32)/2;
          v2_fp32 = v2_fp32 + dt * (kv2_fp32 + mv2_fp32)/2;

          //printf("fp_32 time = %lld\n\r", PT_GET_TIME_usec()-start_time) ;

        // compute DDA in fp16V
        //start_time = PT_GET_TIME_usec();

          kv1_fp16V = v2_fp16V;
          kv2_fp16V = add(mul(k_m_fp16V, v1_fp16V), mul(d_m_fp16V, v2_fp16V)) ;

          v1t_fp16V = add(v1_fp16V, sft(kv1_fp16V, -log_dt)) ;
          v2t_fp16V = add(v2_fp16V, sft(kv2_fp16V, -log_dt)) ;

          mv1_fp16V = v2t_fp16V;
          mv2_fp16V =  add(mul(k_m_fp16V, v1t_fp16V), mul(d_m_fp16V, v2t_fp16V)) ;

          v1_fp16V = add(v1_fp16V, sft(add(kv1_fp16V, mv1_fp16V), -log_dt-1));
          v2_fp16V = add(v2_fp16V, sft(add(kv2_fp16V, mv2_fp16V), -log_dt-1)); 

          //printf("fp_16V time = %lld\n\r", PT_GET_TIME_usec()-start_time) ; 
          
        
        // compute DDA in s15x16
        //start_time = PT_GET_TIME_usec();

          kv1_s15x16 = v2_s15x16;
          kv2_s15x16 = muls15x16(-k_m_s15x16, v1_s15x16) - muls15x16(d_m_s15x16, v2_s15x16) ;
          v1t_s15x16 = v1_s15x16 + (kv1_s15x16>>log_dt) ;//muls15x16(dt_s15x16, kv1_s15x16);
          v2t_s15x16 = v2_s15x16 + (kv2_s15x16>>log_dt) ; //muls15x16(dt_s15x16, kv2_s15x16);
          mv1_s15x16 = v2t_s15x16;
          mv2_s15x16 = muls15x16(-k_m_s15x16, v1t_s15x16) - muls15x16(d_m_s15x16, v2t_s15x16)  ;
          v1_s15x16 = v1_s15x16 + ((kv1_s15x16 + mv1_s15x16)>>(log_dt+1)) ; // muls15x16(dt_s15x16, ((kv1_s15x16 + mv1_s15x16)>>1)) ;
          v2_s15x16 = v2_s15x16 + ((kv2_s15x16 + mv2_s15x16)>>(log_dt+1)) ; 
          
          //s4x27
          kv1_s4x27 = v2_s4x27;
          kv2_s4x27 = muls4x27(-k_m_s4x27, v1_s4x27) - muls4x27(d_m_s4x27, v2_s4x27) ;
          v1t_s4x27 = v1_s4x27 + (kv1_s4x27>>log_dt) ;//muls15x16(dt_s15x16, kv1_s15x16);
          v2t_s4x27 = v2_s4x27 + (kv2_s4x27>>log_dt) ; //muls15x16(dt_s15x16, kv2_s15x16);
          mv1_s4x27 = v2t_s4x27;
          mv2_s4x27 = muls4x27(-k_m_s4x27, v1t_s4x27) - muls4x27(d_m_s4x27, v2t_s4x27)  ;
          v1_s4x27 = v1_s4x27 + ((kv1_s4x27 + mv1_s4x27)>>(log_dt+1)) ; // muls15x16(dt_s15x16, ((kv1_s15x16 + mv1_s15x16)>>1)) ;
          v2_s4x27 = v2_s4x27 + ((kv2_s4x27 + mv2_s4x27)>>(log_dt+1)) ; 
          
          //printf("s15x16 time = %lld\n\r", PT_GET_TIME_usec()-start_time) ; 
          
            
		    } 

        // compute DDA 
  
        // A brief nap
        //new_str = 0;
   }
   PT_END(pt);
} // graphics thread

// ==================================================
// === toggle25 thread on core 0
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    
     // set up LED p25 to blink
     gpio_init(25) ;	
     gpio_set_dir(25, GPIO_OUT) ;
     gpio_put(25, true);
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        // yield time 0.1 second
        //PT_YIELD_usec(100000) ;
        PT_YIELD_INTERVAL(100000) ;

        // toggle the LED on PICO
        LED_state = LED_state? false : true ;
        gpio_put(25, LED_state);
        //
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread

// ==================================================
// === user's serial input thread on core 1
// ==================================================
// serial_read an serial_write do not block any thread
// except this one
static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
      static int test_in1, test_in2, sum ;
      static char cmd[16], arg1[16], arg2[16] ;
      static char* token ;
      //
      while(1) {
        // print prompt
        sprintf(pt_serial_out_buffer, "cmd> ");
        // spawn a thread to do the non-blocking write
        serial_write ;

        // spawn a thread to do the non-blocking serial read
         serial_read ;
        token = strtok(pt_serial_in_buffer, "  ");
        strcpy(cmd, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg1, token) ;
        token = strtok(NULL, "  ");
        strcpy(arg2, token) ;

        
        // rang  (1-10
        if(strcmp(cmd,"k/m")==0){
          sscanf(arg1, "%f", &k_m);
        }

        if(strcmp(cmd,"d/m")==0){
          sscanf(arg1, "%f", &d_m);
        }

        // range  (2-10)
        if(strcmp(cmd,"log_dt")==0){
          sscanf(arg1, "%d", &log_dt);
        }

        // range  (1-10)
        if(strcmp(cmd,"h_scale")==0){
          sscanf(arg1, "%d", &h_scale);
        }

        // range 1-100
        if(strcmp(cmd,"error_gain")==0){
          sscanf(arg1, "%f", &error_gain);
        }

        // values  (fp32, fp16. fix)
        if(strcmp(cmd,"show")==0){
          if(strcmp(arg1,"fp32")==0  | strcmp(arg1,"all")==0) show_fp32 = true ;
          if(strcmp(arg1,"fp16")==0 | strcmp(arg1,"all")==0) show_fp16V = true ;
          if(strcmp(arg1,"fix16")==0 | strcmp(arg1,"all")==0) show_s15x16 = true ;
          if(strcmp(arg1,"fix27")==0 | strcmp(arg1,"all")==0) show_s4x27 = true ;
        }

        // values  (fp32, fp16. fix)
        if(strcmp(cmd,"hide")==0){
          if(strcmp(arg1,"fp32")==0 | strcmp(arg1,"all")==0) show_fp32 = false ;
          if(strcmp(arg1,"fp16")==0 | strcmp(arg1,"all")==0) show_fp16V = false ;
          if(strcmp(arg1,"fix16")==0 | strcmp(arg1,"all")==0) show_s15x16 = false ;
          if(strcmp(arg1,"fix27")==0 | strcmp(arg1,"all")==0) show_s4x27 = false ;
        }

        new_str = true ;

        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // serial thread

// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //
  //  === add threads  ====================
  // for core 1
  //pt_add_thread(protothread_toggle_gpio4) ;
  pt_add_thread(protothread_serial) ;
  //
  // === initalize the scheduler ==========
  pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us
  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.3 two-core, priority\n\r");

  // Initialize the VGA screen
  initVGA() ;
     
  // start core 1 threads
  multicore_reset_core1();
  multicore_launch_core1(&core1_main);

  // === config threads ========================
  // for core 0
  pt_add_thread(protothread_graphics);
  pt_add_thread(protothread_toggle25);
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main